UNIT MMCLoader;
(*
   The MMCLoader handles the MMC access, Filereading, Checking of Checksum and Version and the Flashing for the CPULoader
*)

INTERFACE

{ $W+}                  // enable/disable warnings for this unit

TYPE
  tBootFlag          = Record
                          MinPrevVersion : WORD;                                // Minimum previous version required for the Bootloader to start up and flash
                          Version        : WORD;                                // The version of this file (Build-No)
                          Flag           : WORD;
                       End;

CONST
{!}
  Boot_Start                : LONGWORD  = $3F000;                               // Pointer to the first Byte of the CPULoader-Area: must be adjusted depending your hardware and your FuseBits...
                                                                                // This is a BYTE-address!! (not a WORD address)
  BootFlag[@Boot_Start - 8] : tBootFlag = (MinPrevVersion : $FFFF;              // The minimal required previous Build No. of the APPLICATION - not of the CPULoader!
                                           Version        : $FFFF;              // Do not change these values set here.
                                           Flag           : $FFFF);             // After the BootFlag.Flag another word follows with the checksum of the complete file at [@BOOT_START - 2]
                                                                                // These 3 variables (Version, Flag, CheckSum) must be in the binary file of the application and
                                                                                // thus must be defined there!! They are the last 8 bytes of the APPLICATION memory.
{!}
  Boot_End                  : LongWord  = $40000;                               // Pointer to the first Byte AFTER the End of the Boot Area
  LoaderFlag[@Boot_End - 8] : tBootFlag = (MinPrevVersion : $0000;              // The minimal required previous Build No. of the CPULoader
                                           Version        : PROJECTBUILD+1;     // BuildNo of the BootLoader
                                           Flag           : $BB66);             // Word-Flag, which contains a marker (here $BB66) for the CPULoader.

                                                                                // After the LoaderFlag.Flag another word follows with the checksum of the complete file at [@BOOT_END - 2]
                                                                                // These 4 variables (MinPrevVersion, Version, Flag, CheckSum) must be in the binary file of the application and
                                                                                // thus must be defined there!! They are the last 8 bytes of the Bootloader memory.
                                                                                // ALWAYS check, that the code does NOT overwrite this constant (it is a close shave!)


{!} // -->
  UPDFilename             : STRING  = 'CPU.BIN';                                // This is the name of the Update file. Pls. note: you MUST NOT use long names, just 8.3 convention
  UPDNAME                 : STRING  = 'CPU     ';                               // And this is the name as it is compared here. So, the extension is not used!!
{!} // <--

{$IDATA}

VAR                                                                             // Public variables, procedures and functions of this unit..
{-global variables -----------------------------------------------------------}
   gMinPrevVersion : Word;
   gFileVersion    : Word;
   gFileCheckSum   : Word;


PROCEDURE UpdateFirmware;                                                       // This reads the UpdateFile and flashes it into the application area.
FUNCTION  FATInit                   : BOOLEAN;                                  // Initialize the MMC and check for a valid FAT16 file system
FUNCTION  GetFileVersionAndCheckSum : BOOLEAN;                                  // This reads the UpdateFile, calculates the checksum, compares it with the sum written at the end of the file,
                                                                                // gets the version (PROJECTBUILD Number) in the file and compares it with the version of the current application.

IMPLEMENTATION

CONST
  DOUBLESPEED             : BOOLEAN = FALSE;
  SPIINITCLOCK            : BYTE = %01010011;	// f/128 -> 125 kHz
  SPIREADCLOCK            : BYTE = %01010000;	// f/4   -> 4 MHz
  SPIWRITECLOCK           : BYTE = %01010010;	// f/64  -> 250 kHz

TYPE
  tDirEntry   = RECORD
                  Name    : ARRAY[1..8] OF CHAR;
                  Ext     : ARRAY[1..3] OF CHAR;
                  Attrib  : BYTE;
                  res     : ARRAY[1..10] OF BYTE;
                  Time    : WORD;
                  Date    : WORD;
                  Cluster : WORD;
                  Size    : LONGWORD;
                END;
  tDirPtr     = POINTER TO tDirEntry;       //Klimmzug wegen System.MUL16
  tDirEntries = ARRAY[1..16] OF tDirEntry;

VAR
{$PDATA}
{!}
  MMC_CS  [@PortB, 0] : BIT;                                                    // Check, if this is your hardware SPI Port that is connected to the MMC
  SPIF    [@SPSR, 7]  : Bit;
  
{$IDATA}
{-Buffer und Hilfsvariablen----------------------------------------------------}
  SDDummy           : BYTE;
  buffer            : ARRAY[0..511] OF BYTE;                                    // This BYTE buffer is filled from the Update File
  flash[@buffer]    : ARRAY[0..255] OF WORD;                                    // This is the very same buffer, just addressed as WORDs
  Dir[@buffer]      : tDirEntries;
  FlashAdr          : WORD;
  FlashPage         : BYTE;
  WordsWritten      : LONGWORD;

{-PTI Partition Info-----------------------------------------------------------}
  PTI_StartSector   : LONGWORD;

{-BPB BIOS Parameter Block-----------------------------------------------------}
  BPB_SecPerCluster : BYTE;   //
  BPB_RsvdSecCnt    : WORD;   //Anzahl reserviertet Sektoren
  BPB_NumFATs       : BYTE;
  BPB_FatSz         : WORD;   //Sektorzahl einer FAT

{-FAT--------------------------------------------------------------------------}
  FAT_RootDir       : LONGWORD; //Startsektor des RootDir
  FAT_StartData     : LONGWORD;
  FAT_StartFAT      : LONGWORD;

{-FILE-------------------------------------------------------------------------}
  FirstCluster      : WORD;  //Zeigt auf den aktuellen
  NextCluster       : WORD;  //Zeigt auf den aktuellen oder nchsten Cluster


FUNCTION  FindFirstCluster  : BOOLEAN;             FORWARD;
FUNCTION  FindNextCluster   : BOOLEAN;             FORWARD;
FUNCTION  GetCluster        : BOOLEAN;             FORWARD;
FUNCTION  MMCReadSector(sec : LONGWORD) : BOOLEAN; FORWARD;


CONST
{-BPB BIOS Parameter Block-----------------------------------------------------}
  BPB_BytesPerSec     : WORD = 512;
  SECTOR_SIZE         : WORD = 512;
{-FAT--------------------------------------------------------------------------}
  FAT_RootDirSecs     : WORD = 32;

{$IDATA}

{===============================================================================
  GetFileVersionAndCheckSum
  Checks the Checksum of the file and the Version number of the installed SW
===============================================================================}
Function GetFileVersionAndCheckSum : Boolean;
VAR
  i   : BYTE;
  sec : LONGWORD;
  lCheckSum, lCnt : Word;
  lLCnt : LongWord;
  lBool : Boolean;
  lBootFlagVersion : Word;
BEGIN
  IF NOT FindFirstCluster THEN                                                  // Find the first Cluster of our file: typically a Cluster has 64 sectors, each 512 byte. So, a cluster is 32 kB
     RETURN (FALSE);
  ENDIF;
  lCheckSum := 0;
  lLCnt := 0;
  NextCluster := FirstCluster;                                                  // Set a pointer to the first cluster
  Repeat
     sec := LONGWORD(BPB_SecPerCluster)*LONGWORD(NextCluster-2) + FAT_StartData;
     i := 1;
     While i <= BPB_SecPerCluster DO                                            // and start reading the cluster sector by sector
       If MMCReadSector(sec) Then                                               // MMCReadSector writes a sector to the buffer array.
          lCnt := 0;
          While lCnt <= 511 Do                                                  // now we go through the sector byte by byte
              Inc(lLCnt);                                                       // and count the number of bytes (to know, when we have reached the last sector with the checksum and version!)
              If lLCnt < Boot_Start-1 Then                                      // as long we have not reached the last 2 bytes (i.e. the checksum)
                 lCheckSum := lCheckSum + Word(buffer[lCnt]);                   // we keep adding up the bytes for the checksum
              Else                                                              //
                 If lLCnt = Boot_Start - 1 Then                                 // finally, when we have reached the checksum,
                    gMinPrevVersion := flash[252];                              // we get the required minimum previous Version. We use the WORD addressing of the buffer Array here.
                    gFileVersion    := flash[253];                              // we get the file's Version and the file's Checksum. We use the WORD addressing of the buffer Array here.
                    gFileCheckSum   := flash[255];
                 EndIf;
              EndIf;
              Inc(lCnt);
          EndWhile;
       EndIf;
       sec := sec + 1;
       Inc(i);
     EndWhile;
  Until (Not FindNextCluster);                                                  // we read all clusters in, that are related to the file.
  lBootFlagVersion := BootFlag.Version;                                         // Use a local variable here, the Addressing in the Return(..) Statement will be much shorter!
  Return ((lCheckSum = gFileCheckSum) And                                       // Finally, we return the result of the check: The calculated checksum must be the same as the file's checksum
          (((lBootFlagVersion >= gMinPrevVersion) And                           // AND ( the current applications version must be greater equal than the file's minimum required version
            (gFileVersion > lBootFlagVersion)) Or (lBootFlagVersion = $FFFF))); // AND ( the file's version must be greater than the current version
                                                                                //       OR  the current version must be $FFFF in case of no valid application = erased application flash)
END;


{===============================================================================
  UpdateFirmware
  Fhrt den Update durch
===============================================================================}
PROCEDURE UpdateFirmware;
BEGIN
  FlashInitPage(ADDR(Bootflag));                                                // First thing to do: Erase the page with the Bootflag at the end of the application area
  FlashErasePage;                                                               // If something goes wrong, the CPULoader will always try to repeat the flashing from the MMC.
  FlashAdr  := 0;
  FlashPage := 0;
  WordsWritten := 0;
  NextCluster := FirstCluster;
  GetCluster;
  WHILE FindNextCluster DO
    GetCluster;
  ENDWHILE;
END UpdateFirmware;


{===============================================================================
  SPIPut
  Schreibt ein Byte ber SPI
===============================================================================}
PROCEDURE SPIPut(b: Byte);
BEGIN
  SPDR := b;
  REPEAT UNTIL SPIF;
END SPIPut;

{===============================================================================
  SPIGet
  Holt ein Byte ber SPI
===============================================================================}
FUNCTION SPIGet : BYTE;
BEGIN
  SPDR := $FF;
  REPEAT UNTIL SPIF;
  RETURN (SPDR);
END SPIGet;

{===============================================================================
  MMCSendCmd
  Sendet ein Kommando
===============================================================================}
FUNCTION MMCSendCmd(cmd, argA, argB, argC, argD, CRC : BYTE) : BYTE;
VAR
  result  : BYTE;
  i       : WORD;
BEGIN
  SPIPut($FF);  //Delay
  EXCL(MMC_CS);
  SPIPut(cmd);
  SPIPut(argA);
  SPIPut(argB);
  SPIPut(argC);
  SPIPut(argD);
  SPIPut(CRC);
  i := 1;
  WHILE i <= $1FFF DO
    result := SPIGet;
    IF (result <> $FF) THEN
      Break;
    ENDIF;
    Inc(i);
  ENDWHILE;
  RETURN (result);
END MMCSendCmd;

{===============================================================================
  MMCInit
  Karte initialisieren
===============================================================================}
FUNCTION MMCInit : BOOLEAN;
VAR
  i   : WORD;
  OK  : BOOLEAN;
  res : BYTE;
BEGIN
  INCL(MMC_CS);
{!} // -->                                                                      //  You may have to change the DDRs of Port B here to your processors hardware SPI Port...
  INCL(DDRB, 0);  //CS
  INCL(DDRB, 1);  //SCK ist output
  INCL(DDRB, 2);  //MOSI ist output
  EXCL(DDRB, 0);
  EXCL(DDRB, 1);
  EXCL(DDRB, 2);
  INCL(DDRB, 0);  //CS ist output
  INCL(DDRB, 1);  //SCK ist output
  INCL(DDRB, 2);  //MOSI ist output
  EXCL(DDRB, 3);  //MISO
{!} // <--

  SPCR := SPIINITCLOCK;
  IF DOUBLESPEED THEN
    EXCL(SPSR, 0);
  ELSE
    INCL(SPSR, 0);
  ENDIF;
  i := 1;
  WHILE i <= 10 DO
    SPIGet;
    Inc(i);
  ENDWHILE;
  {---------------------------------------------}
  OK := FALSE;
  i := 1;
  WHILE i <= 512 DO
    res := MMCSendCmd($40, 0, 0, 0, 0, $95); //Reset
    INCL(MMC_CS);
    SPIGet;
    IF res = $01 THEN
      OK := TRUE;
      BREAK;
    ENDIF;
    Inc(i);
  ENDWHILE;
  IF NOT OK THEN
    RETURN (FALSE);
  ENDIF;
  {---------------------------------------------}
  OK := FALSE;
  i := 1;
  WHILE i <= 2048 DO
    res := MMCSendCmd($41, 0, 0, 0, 0, $FF); //Aufwachen
    INCL(MMC_CS);
    SPIGet;
    IF res = $00 THEN
      OK := TRUE;
      BREAK;
    ENDIF;
    Inc(i);
  ENDWHILE;
  RETURN (OK);
  {---------------------------------------------}
  res := MMCSendCmd($50, 0, 0, $02, 0, $FF); //512 Bit Blocklnge
  INCL(MMC_CS);
  OK := (res = $00);
  SPIGet;
  SPIGet;
  RETURN (OK);
END MMCInit;


FUNCTION MMCWaitStart : BOOLEAN;
VAR
  i : INTEGER;
BEGIN
  i := 0;
  WHILE i <= 4096 DO
    IF SPIGet = $FE THEN
      RETURN (TRUE);
    ENDIF;
    Inc(i);
  ENDWHILE;
  RETURN (FALSE);
END MMCWaitStart;

{===============================================================================
  MMCReadSector
  Liest einen Sektor in buffer ein
===============================================================================}
FUNCTION MMCReadSector(sec : LONGWORD) : BOOLEAN;
VAR
  i : Word;
  res : BYTE;
BEGIN
  sec := LONGWORD(SECTOR_SIZE) * (sec + PTI_StartSector);

  SPCR := SPIREADCLOCK;
  IF DOUBLESPEED THEN
    EXCL(SPSR, 0);
  ELSE
    INCL(SPSR, 0);
  ENDIF;

  res := MMCSendCmd( $51,
                    HI(HIWORD(sec)), LO(HIWORD(sec)),
                    HI(LOWORD(sec)), LO(LOWORD(sec)), $FF);

  IF res <> 0 THEN
    INCL(MMC_CS);
    RETURN(FALSE);
  ENDIF;
  IF NOT MMCWaitStart THEN
    INCL(MMC_CS);
    RETURN(FALSE);
  ENDIF;

  i := 0;
  WHILE i <= SECTOR_SIZE - 1 DO
    buffer[i] := SPIGet;
    Inc(i);
  ENDWHILE;
  SPIGet;
  SPIGet;
  INCL(MMC_CS);
  SPIGet;
  RETURN (TRUE);
END MMCReadSector;

{------------------------------------------------------------------------------}

{===============================================================================
  FATInit
  Holt die erste Partition und initialisiert einiges
===============================================================================}
FUNCTION FATInit : BOOLEAN;
VAR
  ww : WORD;
BEGIN
  IF NOT MMCInit THEN
    RETURN (FALSE);
  ENDIF;
  PTI_StartSector := 0;
  IF NOT MMCReadSector(0) THEN
    RETURN (FALSE);
  ENDIF;
  //Begin of the first Partition
  LO(ww) := buffer[$1BE + $08];
  HI(ww) := buffer[$1BE + $09];
  LOWORD(PTI_StartSector) := ww;
  LO(ww) := buffer[$1BE + $0A];
  HI(ww) := buffer[$1BE + $0B];
  HIWORD(PTI_StartSector) := ww;
  //BPB and FAT
  IF NOT MMCReadSector(0) THEN
    RETURN (FALSE);
  ENDIF;
  BPB_SecPerCluster   := buffer[13];
  LO(BPB_RsvdSecCnt)  := buffer[14];
  HI(BPB_RsvdSecCnt)  := buffer[15];
  BPB_NumFATs         := buffer[16];
  LO(BPB_FatSz)       := buffer[22];
  HI(BPB_FatSz)       := buffer[23];
  //FAT
  FAT_RootDir         := LONGWORD(BPB_RsvdSecCnt) + LONGWORD(BPB_FatSz)   * LONGWORD(BPB_NumFATs);
  FAT_StartData       := LONGWORD(BPB_RsvdSecCnt) + LONGWORD(BPB_NumFATs) * LONGWORD(BPB_FatSz) + FAT_RootDirSecs;
  RETURN (TRUE);
END FATInit;

{===============================================================================
  FindFirstCluster
  Sucht im Rootverzeichnis die Datei und liefert den ersten Cluster
  Im Rootverzeichnis sind maximal 512 Eintrge erlaubt. Ein Eintrag ist 32 Byte
  gro, ein Sektor 512 Byte. Somit muss nach 32 Sektoren der Eintrag gefunden
  sein, sofern er vorhanden ist.
===============================================================================}
FUNCTION FindFirstCluster : BOOLEAN;
VAR
  sec : BYTE;
  d   : BYTE;
  p   : tDirPtr;
BEGIN
  sec := 0;
  WHILE sec <= 31 DO
    MMCReadSector(FAT_RootDir + LONGWORD(sec));
    p := @Dir[1];
    d := 1;
    WHILE d <= 16 DO
      WITH p^ DO
        IF (Name[1] = UPDNAME[1]) AND (Name[2] = UPDNAME[2]) AND
           (Name[3] = UPDNAME[3]) AND (Name[4] = UPDNAME[4]) AND
           (Name[5] = UPDNAME[5]) AND (Name[6] = UPDNAME[6]) AND
           (Name[7] = UPDNAME[7]) AND (Name[8] = UPDNAME[8]) THEN
          FirstCluster := Cluster;
          NextCluster  := Cluster;
          RETURN (TRUE);
        ENDIF;
      ENDWITH;
      p++;
      Inc(d);
    ENDWHILE;
    Inc(sec);
  ENDWHILE;
  RETURN(FALSE);
END FindFirstCluster;

{===============================================================================
  FindNextCluster
  Sucht den nchsten Cluster
===============================================================================}
FUNCTION FindNextCluster : BOOLEAN;
VAR
  cluster     : WORD;
  ByteOffset  : LONGWORD;
  SectorOffset: LONGWORD;
BEGIN
  ByteOffset := LONGWORD(NextCluster);
  ByteOffset := ByteOffset * 2;
  SectorOffset := (ByteOffset DIV LONGWORD(BPB_BytesPerSec)) + LONGWORD(BPB_RsvdSecCnt);
  IF MMCReadSector(SectorOffset) THEN
    ByteOffset := ByteOffset AND 511;
    Cluster := LOWORD(ByteOffset);
    LO(NextCluster) := buffer[Cluster];
    HI(NextCluster) := buffer[Cluster + 1];
    IF NextCluster >= $FFF8 THEN //End Of Cluster                               // One Cluster in a 2GB card has 32 kb, so this is the check for the 2 GB limit here
      RETURN(FALSE);
    ELSE
      RETURN(TRUE);
    ENDIF;
  ENDIF;
  RETURN (FALSE);
END FindNextCluster;

{===============================================================================
  WritePage
  Schreibt eine Seite in den Flash
===============================================================================}
PROCEDURE WritePage(idx : BYTE);
VAR
  b : Byte;
BEGIN
  FLASH_ADDR  := FlashAdr;
  FLASH_PAGE  := FlashPage;
  FlashErasePage;
  FLASH_ADDR  := FlashAdr;
  //Load 128 Words / 256 Bytes
  FOR b := idx TO idx + 127 DO                                                  // Do not make this a WHILE..DO Statement! The b-Byte will flow over!
    FlashWritePage(flash[b]);
    INC(WordsWritten);
  ENDFOR;
  FLASH_ADDR   := FlashAdr;
  FlashProgPage;
  INC(FlashAdr, 256);
  IF FlashAdr = 0 THEN
    INC(FlashPage);
  ENDIF;
END WritePage;

{===============================================================================
  GetCluster
  Liest die Daten eines Clusters und programmiert
===============================================================================}
FUNCTION GetCluster : BOOLEAN;
VAR
  i   : BYTE;
  sec : LONGWORD;
BEGIN
  sec := LONGWORD(BPB_SecPerCluster) * LONGWORD((NextCluster-2)) + FAT_StartData;
  i := 1;
  WHILE i <= BPB_SecPerCluster DO
    IF MMCReadSector(sec) AND (WordsWritten < ((Boot_Start Shr 1) - 128)) THEN  // Boot_Start Shr 1   == Boot_Start / 2 == $01F000
      WritePage(0);
      IF WordsWritten >= ((Boot_Start Shr 1) - 1) THEN
        BREAK;
      ENDIF;
      WritePage(128);
      IF WordsWritten >= ((Boot_Start Shr 1) - 1) THEN
        BREAK;
      ENDIF;
    ELSE
      RETURN (FALSE);
    ENDIF;
    sec := sec + 1;
    Inc(i);
  ENDWHILE;
  RETURN (TRUE);
END GetCluster;


INITIALIZATION
END MMCLoader.

